// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "UIAlignableElement.hpp"

using namespace Dawn;

UIAlignableElement::UIAlignableElement() {
  alignUnit[0] = UIAlignmentUnit::SCALE;
  alignUnit[1] = UIAlignmentUnit::SCALE;
  alignUnit[2] = UIAlignmentUnit::SCALE;
  alignUnit[3] = UIAlignmentUnit::SCALE;

  eventAlignmentUpdated = [&](const glm::vec2 p, const glm::vec2 s) {
  };
}

void UIAlignableElement::updateSelfAlignment(
  const glm::vec2 pPos,
  const glm::vec2 pSize,
  const float_t canvasScale
) {
  auto valueAxis = [&](
    const enum UIAlignmentUnit unit,
    const float_t alignment,
    const float_t parentSize,
    const float_t ratioSize,
    const float_t selfSize
  ) {
    if(alignment == UI_ALIGN_AUTO) return selfSize;

    switch(unit) {
      case UIAlignmentUnit::PIXEL:
        return alignment;

      case UIAlignmentUnit::SCALE:
        return canvasScale * alignment;

      case UIAlignmentUnit::PERCENT:
        return parentSize * (alignment / 100.0f);

      case UIAlignmentUnit::RATIO:
        return (alignment / 100.0f) * ratioSize;
      
      default:
        assertUnreachable("Invalid UIAlignmentType");
        return 0.0f;
    }
  };

  auto alignAxis = [&](
    const enum UIAlignmentType type,
    const enum UIAlignmentUnit unit0,
    const enum UIAlignmentUnit unit1,
    const float_t alignment0,
    const float_t alignment1,
    const float_t parentSize,
    const float_t ratioSize,
    const float_t selfSize,
    float_t &outPosition,
    float_t &outSize
  ) {
    switch(type) {
      case UIAlignmentType::START:
        outPosition = valueAxis(
          unit0,
          alignment0,
          parentSize,
          ratioSize,
          selfSize
        );
        outSize = valueAxis(
          unit1,
          alignment1,
          parentSize,
          ratioSize,
          selfSize
        );
        break;

      case UIAlignmentType::MIDDLE:
        outSize = valueAxis(
          unit1,
          alignment1,
          parentSize,
          ratioSize,
          selfSize
        );
        outPosition = (parentSize / 2.0f) - (selfSize / 2.0f) + valueAxis(
          unit0,
          alignment0,
          parentSize,
          ratioSize,
          selfSize
        );
        break;

      case UIAlignmentType::END:
        outSize = valueAxis(
          unit0,
          alignment0,
          parentSize,
          ratioSize,
          selfSize
        );
        outPosition = parentSize - outSize - valueAxis(
          unit1,
          alignment1,
          parentSize,
          ratioSize,
          selfSize
        );
        break;

      case UIAlignmentType::STRETCH:
        outPosition = valueAxis(
          unit0,
          alignment0,
          parentSize,
          ratioSize,
          selfSize
        );
        outSize = parentSize - (outPosition + valueAxis(
          unit1,
          alignment1,
          parentSize,
          ratioSize,
          selfSize
        ));
        break;

      default:
        assertUnreachable("Invalid UIAlignmentType");
    }
  };

  bool_t heightFirst = (
    alignUnit[0] == UIAlignmentUnit::RATIO ||
    alignUnit[2] == UIAlignmentUnit::RATIO
  );

  if(heightFirst) {
    // Align height first, this will define size.y which we can use as the ratio
    // for the width/X axis alignment.
    alignAxis(
      alignY,
      alignUnit[1],
      alignUnit[3],
      align[1],
      align[3],
      pSize.y,
      0,
      this->getSelfHeight(),
      position.y,
      size.y
    );
    alignAxis(
      alignX,
      alignUnit[0],
      alignUnit[2],
      align[0],
      align[2],
      pSize.x,
      size.y,
      this->getSelfWidth(),
      position.x,
      size.x
    );
  } else {
    alignAxis(
      alignX,
      alignUnit[0],
      alignUnit[2],
      align[0],
      align[2],
      pSize.x,
      0,
      this->getSelfWidth(),
      position.x,
      size.x
    );
    alignAxis(
      alignY,
      alignUnit[1],
      alignUnit[3],
      align[1],
      align[3],
      pSize.y,
      size.x,
      this->getSelfHeight(),
      position.y,
      size.y
    );
  }

  this->position += pPos;
  this->eventAlignmentUpdated(position, size);
}

float_t UIAlignableElement::getSelfWidth() {
  return size.x;
}

float_t UIAlignableElement::getSelfHeight() {
  return size.y;
}

void UIAlignableElement::updateAlignment(
  const glm::vec2 pPos,
  const glm::vec2 pSize,
  const float_t canvasScale
) {
  this->updateSelfAlignment(pPos, pSize, canvasScale);

  // Now update children alignment
  auto children = getChildren();
  for(auto &c : children) {
    c->updateAlignment(this->position, this->size, canvasScale);
  }
}